from trio_websocket import ConnectionClosed

from . import BaseTransport, TransportClosed


class WebSocketTransport(BaseTransport):
    def __init__(self, ws):
        self._ws = ws

    async def recv(self) -> bytes:
        try:
            return await self._ws.get_message()
        except ConnectionClosed:
            raise TransportClosed()

    async def send(self, data: bytes) -> None:
        try:
            return await self._ws.send_message(data)
        except ConnectionClosed:
            raise TransportClosed()



from contextlib import asynccontextmanager
import enum
import json
import logging
import ssl
import typing

from sansio_jsonrpc import (
    JsonRpcException,
    JsonRpcInternalError,
    JsonRpcPeer,
    JsonRpcRequest,
    JsonRpcResponse,
)
import trio
import trio_websocket

from .transport import BaseTransport, TransportClosed
from .transport.memory import MemoryTransport
from .transport.ws import WebSocketTransport


logger = logging.getLogger("trio_jsonrpc")


class JsonRpcConnectionType(enum.Enum):
    """
    An enumeration that identifies whether the peer is a client role or a server role.
    """

    CLIENT = 0
    SERVER = 1


class JsonRpcConnection:
    """ A JSON-RPC client. """

    def __init__(self, transport, peer_type):
        """ Constructor. """
        self._transport = transport
        self._peer_type = peer_type
        self._sansio_peer = JsonRpcPeer()
        self._bg_task_running = False
        self._outbound_requests = dict()
        irsend, irrecv = trio.open_memory_channel(0)
        self._inbound_requests_send = irsend
        self._inbound_requests_recv = irrecv

    @property
    def is_server(self):
        """ Returns True if this peer is in the server role. """
        return self._peer_type == JsonRpcConnectionType.SERVER

    @property
    def is_client(self):
        """ Returns True if this peer is in the client role. """
        return self._peer_type == JsonRpcConnectionType.CLIENT

    async def request(
        self, method: str, params: typing.Union[dict, list] = None
    ) -> typing.Any:
        """
        Send a request to the server and return its result.

        :returns: a response from the server
        :raises: a subclass of class:`JsonRpcException` if the server returns an error
        """
        request_id, bytes_to_send = self._sansio_peer.request(
            method=method, params=params
        )
        # The background task provides a response to this task using a one-time channel.
        response_send, response_recv = trio.open_memory_channel(0)
        self._outbound_requests[request_id] = response_send
        await self._transport.send(bytes_to_send)
        response = await response_recv.receive()
        if response.success:
            return response.result
        else:
            raise JsonRpcException.exc_from_error(response.error)

    async def notify(
        self, method: str, params: typing.Union[dict, list] = None
    ) -> None:
        """
        Send a notification.

        This does expect or wait for any response.
        """
        bytes_to_send = self._sansio_peer.notify(method, params)
        await self._transport.send(bytes_to_send)

    async def iter_requests(self):
        """
        An asynchronous iterator that yields each request (including notifications) as
        it is received.

        This is intended to be called on server objects, but this restriction is not
        enforced.
        """
        async for request in self._inbound_requests_recv:
            yield request

    async def respond_with_result(self, request, result):
        bytes_to_send = self._sansio_peer.respond_with_result(request, result)
        await self._transport.send(bytes_to_send)

    async def respond_with_error(self, request, error):
        bytes_to_send = self._sansio_peer.respond_with_error(request, error)
        await self._transport.send(bytes_to_send)

    async def _background_task(self):
        """
        The background task handles incoming messages.
        """
        self._bg_task_running = True

        while self._bg_task_running:
            try:
                bytes_received = await self._transport.recv()
                messages = self._sansio_peer.parse(bytes_received)
                for message in messages:
                    # sansio-jsonrpc guarantees that each message is either a request
                    # or a response.
                    if isinstance(message, JsonRpcRequest):
                        await self._inbound_requests_send.send(message)
                    else:
                        assert isinstance(message, JsonRpcResponse)
                        try:
                            response_send = self._outbound_requests.pop(message.id)
                            await response_send.send(message)
                        except KeyError:
                            id_ = message.id
                            msg = f"No in-flight request matches response.id={id_}"
                            logger.error(msg)
                            await self._background_send_error(JsonRpcInternalError(msg))
            except JsonRpcException as jre:
                if self.is_client:
                    # As client, we don't need to send a response, so we just log the
                    # error.
                    logger.exception("JSON-RPC exception in client background task.")
                else:
                    # As server, we should try to send an error response.
                    logger.exception("JSON-RPC exception in server background task.")
                    await self._background_send_error(jre)
            except trio.Cancelled:
                # If cancelled, end the loop.
                break
            except TransportClosed:
                # If the transport is closed on the receive side, we need to exit the
                # loop.
                logger.info(
                    "Background task is exiting because the receive transport is closed."
                )
                # We also close our requests channel so that any callers inside
                # `iter_requests()` will move on.
                await self._inbound_requests_send.aclose()
                break
            except Exception:
                # An uncaught exception shouldn't crash the background task, but we also
                # don't have any useful handling we can perform here.
                logger.exception("Unhandled exception in JSON-RPC background task.")

        self._bg_task_running = False

    async def _background_send_error(self, exc, request=None):
        try:
            bytes_to_send = self._sansio_peer.respond_with_error(
                request, exc.get_error()
            )
            await self._transport.send(bytes_to_send)
        except TransportClosed:
            # If the transport is closed on the send() side, then we keep the loop
            # running in case the transport is half-closed and we might still be able to
            # receive data.
            logger.error(
                "Server cannot send error response because the transport is closed."
            )


def jsonrpc_client(
    transport: BaseTransport, nursery: trio.Nursery,
) -> JsonRpcConnection:
    """ Create a JSON-RPC peer instance using the specified transport. """
    peer = JsonRpcConnection(transport, JsonRpcConnectionType.CLIENT)
    nursery.start_soon(peer._background_task)
    return peer


def jsonrpc_server(
    transport: BaseTransport, nursery: trio.Nursery, request_buffer_len: int = 1,
) -> JsonRpcConnection:
    """ Create a JSON-RPC peer instance using the specified transport. """
    request_buffer_send, request_buffer_recv = trio.open_memory_channel(
        request_buffer_len
    )

    peer = JsonRpcConnection(transport, JsonRpcConnectionType.SERVER)
    nursery.start_soon(peer._background_task)
    return peer


@asynccontextmanager
async def open_jsonrpc_memory(
    send_channel: trio.abc.SendChannel, recv_channel: trio.abc.ReceiveChannel,
) -> typing.AsyncIterator[JsonRpcConnection]:
    """
    Open a JSON-RPC connection using Trio channels as transport.

    This is mainly intended for testing, since the client and server must be running
    inside the same process.
    """
    async with trio.open_nursery() as nursery:
        transport = MemoryTransport(send_channel, recv_channel)
        yield jsonrpc_client(transport, nursery)
        nursery.cancel_scope.cancel()


@asynccontextmanager
async def serve_jsonrpc_memory(
    send_channel: trio.abc.SendChannel, recv_channel: trio.abc.ReceiveChannel,
):
    """
    Serve a JSON-RPC connection using Trio channels as transport.

    This is mainly intended for testing, since the client and server must be running
    inside the same process. Note that this only accepts 1 "connection": the one passed
    to this function.
    """
    async with trio.open_nursery() as nursery:
        transport = MemoryTransport(send_channel, recv_channel)
        yield jsonrpc_server(transport, nursery)
        nursery.cancel_scope.cancel()


@asynccontextmanager
async def open_jsonrpc_ws(url: str) -> typing.AsyncIterator[JsonRpcConnection]:
    """ Open a JSON-RPC connection using WebSocket transport. """
    async with trio_websocket.open_websocket_url(url) as ws:
        async with trio.open_nursery() as nursery:
            transport = WebSocketTransport(ws)
            yield jsonrpc_client(transport, nursery)
            nursery.cancel_scope.cancel()



"""
This module contains a flexible dispatch system for routing requests to a JSON-RPC
server to specific handler functions. Use of this module is completely optional, and it
is entirely possible to dispatch JSON-RPC methods yourself by directly calling the
server's ``iter_requests()`` method.
"""
from copy import copy
from functools import partial
import logging
import types
import typing

from sansio_jsonrpc import JsonRpcRequest
import trio
from trio_jsonrpc import (
    JsonRpcConnection,
    JsonRpcInternalError,
    JsonRpcException,
    JsonRpcMethodNotFoundError,
)


logger = logging.getLogger(__name__)


class Dispatch:
    """
    This class assists with dispatching JSON-RPC methods to specific handler functions.

    Each handler is registered using a decorator. When a method is executed on the
    dispatcher, it looks up the registered handler and calls it in a new task.
    """

    def __init__(self):
        """
        Constructor.

        :param nursery: A Nursery to launch handler tasks in.
        """
        self._handlers = dict()

    def handler(self, fn):
        """
        A decorator that registers an async function as a handler.

        :param fn: The function to decorate.
        """
        try:
            name = fn.__name__
        except AttributeError:
            raise RuntimeError(
                "The Dispatch.handler() decorator must be applied to a named function."
            )
        self._handlers[name] = fn

    async def execute(
        self,
        request: JsonRpcRequest,
        result_channel: trio.MemorySendChannel,
        context: typing.Any = None,
    ) -> typing.Any:
        """
        Dispatch a JSON-RPC request and send its result to the given channel.

        :param request:
        :param result_channel:
        :param context:
        """
        try:
            handler = self._get_handler(request.method)
            params = request.params
            if context is None:
                if isinstance(params, list):
                    fn = partial(handler, *params)
                elif isinstance(params, dict):
                    fn = partial(handler, **params)
                else:
                    fn = handler
            else:
                if isinstance(params, list):
                    fn = partial(handler, context, *params)
                elif isinstance(params, dict):
                    fn = partial(handler, context, **params)
                else:
                    fn = partial(handler, context)
            result = await fn()
        except JsonRpcException as jre:
            result = jre
        except Exception as exc:
            logger.exception(
                'An unhandled exception occurred in handler "%s" context=%r',
                handler.__name__,
                context,
            )
            result = JsonRpcInternalError("An unhandled exception occurred.")
        await result_channel.send((request, result))

    def _get_handler(self, method: str):
        """ Find the handler function for a given JSON-RPC method name. """
        try:
            return self._handlers[method]
        except KeyError:
            raise JsonRpcMethodNotFoundError(f'Method "{method}" not found.') from None



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['trio_jsonrpc', 'trio_jsonrpc.transport']

package_data = \
{'': ['*']}

install_requires = \
['sansio-jsonrpc==0.1.0', 'trio-websocket>=0.8.0,<0.9.0']

setup_kwargs = {
    'name': 'trio-jsonrpc',
    'version': '0.1.0',
    'description': 'JSON-RPC v2.0 for Trio',
    'long_description': '# JSON-RPC v2.0 for Trio\n\nThis project provides an implementation of [JSON-RPC v\n2.0](https://www.jsonrpc.org/specification) based on\n[sansio-jsonrpc](https://github.com/hyperiongray/sansio-jsonrpc) with all of the I/O\nimplemented using the [Trio asynchronous framework](https://trio.readthedocs.io).\n\n## Client Example\n\nThe following example shows a basic JSON-RPC client.\n\n```python\nfrom trio_jsonrpc import open_jsonrpc_ws, JsonRpcException\n\nasync def main():\n    async with open_jsonrpc_ws(\'ws://example.com/\') as client:\n        try:\n            resp = await client.request(\n                method=\'open_vault_door\',\n                {\'employee\': \'Mark\', \'pin\': 1234}\n            )\n            print(resp.result)\n\n            await client.notify(method=\'hello_world\')\n        except JsonRpcException as jre:\n            print(\'RPC failed:\', jre)\n\ntrio.run(main)\n```\n\nThe example begins by opening a JSON-RPC connection using a WebSocket transport. The\nimplementation is designed to support multiple types of transport, but currently\nWebSocket transport is the only one that has been implemented.\n\n> Note that JSON-RPC does not contain any framing logic, i.e. a specification for how to\n> identify message boundaries within a stream. Therefore, if you want to use JSON-RPC\n> with raw TCP sockets, you either need to add your own framing logic or else use a\n> streaming JSON parser. For this reason, we have chosen to focus on WebSocket transport\n> initially, because WebSocket does include framing.\n\nThe connection is opened inside a context manager that guarantees the connection is\nready when entering the block and automatically closes the connection when leaving the\nblock.\n\nWithin the block, we call the client\'s `request(...)` method to send a JSON-RPC request.\nThis method sends the request to the server, waits for a response, and returns a\n`JsonRpcResponse` object. If the server indicates that an error occurred, a\n`JsonRpcException` will be raised instead. The client multiplexes requests so that it\ncan be use concurrently from multiple tasks; responses are routed back to the\nappropriate task that called `request(...)`.\n\nThe client also has a `notify(...)` method which sends a request to the server but does\nnot expect or wait for a response.\n\n## Server Example\n\nThe following example shows a basic JSON-RPC server. The server is more DIY (do it\nyourself) than the client because a server has to incorporate several disparate\nfunctionalities:\n\n1. Setting up the transport, especially if the transport requires a handshake as\n   WebSocket does.\n2. Handling new connections to the server.\n3. Multiplexing requests on a single connection.\n4. Dispatching a request to an appropriate handler.\n5. Managing connection state over the course of multiple requests. (I.e. allowing one\n   handler to indicate that the connection is authorized, so other handlers can use that\n   authorization information to make access control decisions.)\n6. Possibly logging information about each request.\n\nThis library cannot feasibly implement a default solution that handles the\naforementioned items in a way that satsifies every downstream project. Instead, the\nlibrary gives you the pieces you need to build a server.\n\n```python\nfrom dataclasses import dataclass\nimport trio\nfrom trio_jsonrpc import Dispatch, JsonRpcApplicationError\nimport trio_websocket\n\n@dataclass\nclass RequestContext:\n    """ A sample implementation for request context. """\n    db: typing.Any = None\n    authorized_employee: str = None\n\ndispatch = Dispatch()\n\n@dispatch.handler\nasync def open_vault_door(context, employee, pin):\n    access = await context.db.check_pin(employee, pin)\n    if access:\n        context.authorized_employee = employee\n        return {"door_open": True}\n    else:\n        context.authorized_employee = None\n        raise JsonRpcApplicationError(code=-1, message="Not authorized.")\n\n@dispatch.handler\nasync def close_vault_door(context):\n    context.authorized_employee = None\n    return {"door_open": False}\n\nasync def main():\n    db = ...\n    base_context = RequestContext(db=db)\n\n    async def responder(conn, recv_channel):\n        async for result in recv_channel:\n            if\n\n    async def connection_handler(ws_request):\n        ws = await ws_request.accept()\n        transport = WebSocketTransport(ws)\n        rpc_conn = JsonRpcConnection(transport, JsonRpcConnectionType.SERVER)\n        conn_context = copy(base_context)\n        async with trio.open_nursery() as nursery:\n            nursery.start_soon(rpc_conn._background_task)\n            async for request in rpc_conn.iter_requests():\n                nursery.start_soon(dispatch, request, conn_context)\n            nursery.cancel_scope.cancel()\n\n    await trio_websocket.serve_websocket(connection_handler, \'localhost\', 8000, None)\n\ntrio.run(main)\n```\n\nThis example defines a `RequestContext` class which is used to share state between\nrequests on the same connection. Next, a `Dispatch` object is created, which is used to\nmap JSON-RPC methods to Python functions. The `@dispatch.handler` decorator\nautomatically registers a Python function as a JSON-RPC method of the same name. Each\nof these handlers takes a `context` object as well as the parameters included in the\nJSON-RPC request. The use of the dispatch and/or context systems are entirely optional.\n\nIn `main()`, we set up a new WebSocket server. For each new connection, we complete the\nWebSocket handshake and then wrap the connection in a `JsonRpcConnection`. Finally, we\niterate over the incoming JSON-RPC requsts and dispatch each one inside a new task.\n',
    'author': 'Mark E. Haase',
    'author_email': 'mehaase@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/hyperiongray/trio-jsonrpc',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABC, abstractmethod


class BaseTransport(ABC):
    """ A base class for JSON-RPC transports. """

    @abstractmethod
    async def recv(self) -> bytes:
        """ Receive data from the transport. """

    @abstractmethod
    async def send(self, data: bytes):
        """ Send data through the transport."""


class TransportClosed(Exception):
    pass
